/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <PlatformOS.h>
#include <JSContextRefPrivate.h>
#include <JSSystemEventDispatcher.h>
#include <JSExtensionData.h>
#include "JSExtensionUtils.h"

namespace JSExtUtils
{
	void replaceSubstring(std::string & str, std::string target, std::string replacement)
	{
		size_t index = str.find(target, index); 
		while (index != std::string::npos) { 
			 
			 /* Make the replacement. */ 
			 str.replace(index, target.length(), replacement); 
 
			 /* Advance index forward so the next iteration doesn't pick it up as well. */ 
			 index += replacement.length(); 
			 index = str.find(target, index); 
		} 
	}

	bool BlobToCharArray(WebCore::Blob* targetBlob, char** outputArray, unsigned* outputArraySize)
	{
		if(!outputArray || !outputArraySize)
		{
			return false;
		}
		//Access BlobStorageData from the registry, this is only supposed to happen on the MainThread
		WebCore::BlobRegistryImpl& blobReg = (WebCore::BlobRegistryImpl&)WebCore::blobRegistry();
		PassRefPtr<WebCore::BlobStorageData> blobStorageData = blobReg.getBlobDataFromURL(targetBlob->url());

		if(*outputArray)
		{
			delete [] (*outputArray);
		}

		*outputArraySize = 0;

		for (WebCore::BlobDataItemList::const_iterator iter = blobStorageData->items().begin(); iter != blobStorageData->items().end(); ++iter) {
			switch (iter->type) {
			case WebCore::BlobDataItem::Data:
				*outputArraySize += iter->data->length();
				break;
			case WebCore::BlobDataItem::File:
				break;
			case WebCore::BlobDataItem::Blob:
				break;
			}
		}

		if(*outputArraySize < 0)
		{
			return false;
		}

		(*outputArray) = new char[*outputArraySize];
		unsigned dataIndex = 0;
		for (WebCore::BlobDataItemList::const_iterator iter = blobStorageData->items().begin(); iter != blobStorageData->items().end(); ++iter) {
			switch (iter->type) {
			case WebCore::BlobDataItem::Data:
				memcpy((*outputArray) + dataIndex, iter->data->data(), iter->data->length());
				dataIndex += iter->data->length();
				break;
			case WebCore::BlobDataItem::File:
				break;
			case WebCore::BlobDataItem::Blob:
				break;
			}
		}

		return true;
	}

	JSObjectRef DataToJSBlobRef(JSContextRef context, char* data, size_t dataSize)
	{
		PassRefPtr<WebCore::RawData> rawData = WebCore::RawData::create();
		rawData->mutableData()->append(data, dataSize);

		PassOwnPtr<WebCore::BlobData> blobData = WebCore::BlobData::create();
		String contentType;
		contentType.append(L"text/html", 9);
		blobData->setContentType(contentType);
		blobData->appendData(rawData, 0, dataSize);

		PassRefPtr<WebCore::Blob> blob = WebCore::Blob::create(blobData, dataSize);
	
		JSC::ExecState* exec = toJS(context);

		JSC::JSValue jsBlob = WebCore::toJS(exec, static_cast<WebCore::JSDOMGlobalObject*>(exec->lexicalGlobalObject()), blob.get());

		return JSValueToObject(context, toRef(exec, jsBlob), NULL);
	}

	WebCore::Blob* JSObjectRefToBlob(JSContextRef context, JSValueRef blobObjectRef)
	{
		JSC::ExecState* exec = toJS(context);
		JSC::JSValue jsValue = toJS(exec, blobObjectRef);
		//Confirm that the converted object is a Blob
		if(!jsValue.getObject()->inherits(&WebCore::JSBlob::s_info))
		{
			return NULL;
		}
		return WebCore::toBlob(jsValue);
	}

	void* JSObjectRefToNativeImagePointer(JSContextRef context, JSValueRef imageObjectRef)
	{
		// Convert the argument into an actual js object
		JSObjectRef object_reference = JSValueToObject(context, imageObjectRef, 0);
		JSC::JSObject *js_object = toJS(object_reference);

		// Make sure the object passed in is the correct type
		if(!js_object->inherits(&WebCore::JSHTMLImageElement::s_info)) return NULL;
		
		// Drill down into the image element to retrieve its cached image
		WebCore::JSHTMLImageElement *js_html_image_element = static_cast<WebCore::JSHTMLImageElement *>(asObject(js_object));
		WebCore::HTMLImageElement *html_image_element = static_cast<WebCore::HTMLImageElement *>(js_html_image_element->impl());
		WebCore::CachedImage *cached_image = html_image_element->cachedImage();

		if(!cached_image)
		{
			OS::Report("Image has no source.\n");
			return NULL;
		}

		// Retrieve the native texture
		void *native_texture = cached_image->image()->getNativeTexture();
		if(!native_texture)
		{
			OS::Report("Image hasn't finished loading.\n");
		}
		return native_texture;
	}

	bool NativeImageToJPEGAsCharArray(unsigned char* rgbaBigEndianData, unsigned width, unsigned height, unsigned pitch, char** outputArray, unsigned* outputArraySize, WebCore::ColorByteIndices indices)
	{
		Vector<char> convertedDataBuffer;
		WebCore::IntSize size;
		size.setWidth(width);
		size.setHeight(height);
		//Convert the Image pointer into a JPEG
		if(!WebCore::compress32BitImageToJPEGWithPitch(rgbaBigEndianData, size, convertedDataBuffer, pitch, indices))
		{
			return false;
		}

		if(*outputArray)
		{
			delete [] (*outputArray);
		}

		*outputArraySize = convertedDataBuffer.size();
		*outputArray = new char[*outputArraySize];

		memcpy(*outputArray, convertedDataBuffer.data(), *outputArraySize);

		return true;
	}

	JSValueRef JSObjectSetReadOnlyNumberProperty(JSContextRef context, JSObjectRef baseObject, char* propertyName, double propertyValue)
	{
		JSStringRef propNameRef = JSStringCreateWithUTF8CString(propertyName);
		JSValueRef propValRef = JSValueMakeNumber(context, propertyValue);
		JSObjectSetProperty(context, baseObject, propNameRef, propValRef ,kJSPropertyAttributeReadOnly, NULL);
		JSStringRelease(propNameRef);

		return propValRef;
	}

	JSValueRef JSObjectSetReadOnlyUTF8CStringProperty(JSContextRef context, JSObjectRef baseObject, char* propertyName, char* propertyValue)
	{
		JSStringRef propNameRef = JSStringCreateWithUTF8CString(propertyName);
		JSStringRef propValString = JSStringCreateWithUTF8CString(propertyValue);
		JSValueRef propValRef = JSValueMakeString(context, propValString);
		JSObjectSetProperty(context, baseObject, propNameRef, propValRef ,kJSPropertyAttributeReadOnly, NULL);
		JSStringRelease(propNameRef);
		JSStringRelease(propValString);

		return propValRef;
	}

	bool checkCallbackScope(JSContextRef callbackContext, JSValueRef callbackScope, JSClassRef expectedClass, JSValueRef* exception)
	{
		if(!JSValueIsObjectOfClass(callbackContext, callbackScope, expectedClass))
		{
			if(exception)
			{
				JSGlobalContextRef globalContextRef = JSContextGetGlobalContext(callbackContext);
				JSStringRef errorString;
				char infoBuffer[2048];
				JSStringRef infoJSStr = JSContextCreateBacktrace(globalContextRef, 2);
				JSStringGetUTF8CString(infoJSStr, infoBuffer, 2048);
				JSStringRelease(infoJSStr);

				char errorBuffer[2048];
				sprintf(errorBuffer, "%s : Cannot access, scope is invalid\n", infoBuffer);
				LibPlatform::OS::Report(errorBuffer);
				sprintf(errorBuffer, "%s : Cannot access, scope is invalid    ", infoBuffer);
				
				createErrorObjectWithUTF8String(callbackContext, errorBuffer, exception);
			}
			return false;
		}
		return true;
	}

	std::string getJSTypeAsString(JSType type)
	{
		std::string retVal = "undefined";
		switch(type)
		{
		case kJSTypeNull:
			retVal = "null";
			break;
		case kJSTypeBoolean:
			retVal = "Boolean";
			break;
		case kJSTypeNumber:
			retVal = "Number";
			break;
		case kJSTypeString:
			retVal = "String";
			break;
		case kJSTypeObject:
			retVal = "Object";
			break;

		}
		return retVal;
	}

	JSObjectRef createErrorObjectWithUTF8String(JSContextRef context, char* errorString, JSValueRef* exception)
	{
		LibPlatform::OS::Report(errorString);
		JSStringRef errorJSStr = JSStringCreateWithUTF8CString(errorString);
		JSValueRef error = JSValueMakeString(context, errorJSStr);
		JSStringRelease(errorJSStr);
		JSObjectRef errorObj = JSObjectMakeError(context, 1, &error, exception);
		if(exception)
		{
			*exception = errorObj;
		}
		return errorObj;
	}

	bool checkArgumentList(size_t expectedArgumentCount, JSType expectedTypes[], JSContextRef context,  size_t providedArgumentCount, const JSValueRef arguments[], JSObjectRef thisObject, JSValueRef* exception, size_t numDefaultArguments)
	{
		JSStringRef errorString;
		if(thisObject && !JSObjectGetPrivate(thisObject))
		{
			if(exception)
			{
				JSGlobalContextRef globalContextRef = JSContextGetGlobalContext(context);
				char infoBuffer[2048];
				JSStringRef infoJSStr = JSContextCreateBacktrace(context, 2);
				JSStringGetUTF8CString(infoJSStr, infoBuffer, 2048);
				JSStringRelease(infoJSStr);
				char errorBuffer[2048];
				sprintf(errorBuffer, "%s: Invalid scope for callback function.", infoBuffer);
				createErrorObjectWithUTF8String(context, errorBuffer, exception);
			}
			return false;
		}
	
		//confirm that the number of arguments doesn't exceed the maximum and are at least the minimum
		if(expectedArgumentCount >= providedArgumentCount)
		{
			if(providedArgumentCount >= (expectedArgumentCount - numDefaultArguments))
			{
				JSType type;
				//check each argument against the type provided
				for(int i = 0; i < providedArgumentCount; ++i)
				{
					type = JSValueGetType(context, arguments[i]);
					//if the types of the parameters don't match
					if(type != expectedTypes[i])
					{
						if(exception)
						{
							JSGlobalContextRef globalContextRef = JSContextGetGlobalContext(context);
							std::string providedType = getJSTypeAsString(type);
							std::string expectedType = getJSTypeAsString(expectedTypes[i]);
							char infoBuffer[2048];
							JSStringRef infoJSStr = JSContextCreateBacktrace(context, 2);
							JSStringGetUTF8CString(infoJSStr, infoBuffer, 2048);
							JSStringRelease(infoJSStr);
							char errorBuffer[2048];
							sprintf(errorBuffer, "%s \nParameter %u is of type %s, not type %s as expected\n", infoBuffer, i + 1, providedType.c_str(), expectedType.c_str());
							LibPlatform::OS::Report(errorBuffer);
							sprintf(errorBuffer, "%s :Parameter %u is of type %s, not type %s as expected    ", infoBuffer, i + 1, providedType.c_str(), expectedType.c_str());
							
							createErrorObjectWithUTF8String(context, errorBuffer, exception);
						}

						return false;
					}
				}
				return true;
			}
			else
			{
				if(exception)
				{
					JSGlobalContextRef globalContextRef = JSContextGetGlobalContext(context);
					char infoBuffer[2048];
					JSStringRef infoJSStr = JSContextCreateBacktrace(context, 2);
					JSStringGetUTF8CString(infoJSStr, infoBuffer, 2048);
					JSStringRelease(infoJSStr);
					char errorBuffer[2048];
					sprintf(errorBuffer, "%s \nThe number of provided parameters (%u) was less than the minimum number (%u)\n", infoBuffer, providedArgumentCount, (expectedArgumentCount - numDefaultArguments));
					LibPlatform::OS::Report(errorBuffer);
					sprintf(errorBuffer, "%s: The number of provided parameters (%u) was less than the minimum number (%u)    ", infoBuffer, providedArgumentCount, (expectedArgumentCount - numDefaultArguments));
					
					createErrorObjectWithUTF8String(context, errorBuffer, exception);
				}
			}
			return false;
		}


		//if an exception pointer was passed
		if(exception)
		{
			JSGlobalContextRef globalContextRef = JSContextGetGlobalContext(context);
			char infoBuffer[2048];
			JSStringRef infoJSStr = JSContextCreateBacktrace(context, 2);
			JSStringGetUTF8CString(infoJSStr, infoBuffer, 2048);
			JSStringRelease(infoJSStr);
			char errorBuffer[2048];
			sprintf(errorBuffer, "%s \nThe number of provided parameters (%u) exceeded the expected number (%u)\n", infoBuffer, providedArgumentCount, expectedArgumentCount);
			LibPlatform::OS::Report(errorBuffer);
			sprintf(errorBuffer, "%s: The number of provided parameters (%u) exceeded the expected number (%u)    ", infoBuffer, providedArgumentCount, expectedArgumentCount);
			
			createErrorObjectWithUTF8String(context, errorBuffer, exception);
		}
	
		return false;
	}

	static JSClassRef theClass = 0;

	static void packageFinalize(JSObjectRef object)
	{
		/* Finalizes*/
		std::string* packageName = (std::string*)JSObjectGetPrivate(object);
		if(packageName)
		{
			OS::Report("Destroying Package: %s \n", packageName->c_str());
			delete packageName;
			JSObjectSetPrivate(object, NULL);
		}
	}

	static JSClassDefinition packageClassDef = {
		0,             /* version */
		0,             /* attributes */

		"Package",        /* name */
		0,             /* parent class */

		0,             /* static values */
		0,			   /* static functions */

		0,             /* initialize callback */
		packageFinalize,             /* finalize callback */
		0,             /* hasProperty callback */
		0,             /* getProperty callback */
		0,             /* setProperty callback */
		0,             /* deleteProperty callback */
		0,             /* getPropertyNames callback */
		0,             /* callAsFunction callback */
		0,             /* callAsConstructor callback */
		0,             /* hasInstance callback */
		0              /* convertToType callback */
	};



	JSObjectRef getPackageObject(JSContextRef ctx, JSObjectRef base, std::string packageName)
	{
		if(!theClass)
		{
			theClass = JSClassCreate(&packageClassDef);
		}
		unsigned int prevIndex = 0;
		unsigned int index = packageName.find(".");
		std::string subString;
		std::string curPackage = "";
		JSObjectRef prevPackage = base;
		JSObjectRef package;
		JSStringRef jsPackageStr;
		JSType typeCheck;
		JSValueRef packageVal;
		while(index != std::string::npos)
		{
			subString = packageName.substr(prevIndex, index);
			jsPackageStr = JSStringCreateWithUTF8CString(subString.data());
			curPackage.append(subString);
			packageVal = JSObjectGetProperty(ctx, prevPackage, jsPackageStr, NULL);
			typeCheck = JSValueGetType(ctx, packageVal);
			//if the subPackage doesn't exist
			if(typeCheck == kJSTypeUndefined)
			{
				//create it
				std::string* privData = new std::string(curPackage);
				package = JSObjectMake(ctx, theClass, privData);
				JSObjectSetProperty(ctx, prevPackage, jsPackageStr, package, kJSPropertyAttributeNone, NULL);
			}
			else
			{
				package = JSValueToObject(ctx, packageVal, NULL);
			}

			JSStringRelease(jsPackageStr);
			prevPackage = package;
			packageName = packageName.substr(index + 1, packageName.length() - index);
			curPackage.append(".");
			index = packageName.find(".");
		}

		subString = packageName;
		jsPackageStr = JSStringCreateWithUTF8CString(subString.data());

		packageVal = JSObjectGetProperty(ctx, prevPackage, jsPackageStr, NULL);
		typeCheck = JSValueGetType(ctx, packageVal);
		//if the subPackage doesn't exist
		if(typeCheck == kJSTypeUndefined)
		{
			curPackage.append(subString);
			//create it
			std::string* privData = new std::string(curPackage);
			package = JSObjectMake(ctx, theClass, privData);
			JSObjectSetProperty(ctx, prevPackage, jsPackageStr, package, kJSPropertyAttributeNone, NULL);
		}
		else
		{
			package = JSValueToObject(ctx, packageVal, NULL);
		}
		JSStringRelease(jsPackageStr);

		return package;
	}

	bool deleteJSObjectRefByString(JSContextRef ctx, JSObjectRef base, std::string targetName)
	{
		JSStringRef jsPackageStr = JSStringCreateWithUTF8CString(targetName.data());
		bool deleteResult = JSObjectDeleteProperty(ctx, base, jsPackageStr, NULL);
		JSStringRelease(jsPackageStr);
		return deleteResult;
	}

	JSValueRef JSObjectGetPropertyByUTF8CString(JSContextRef context, JSObjectRef base, char* targetName, JSValueRef* exception)
	{
		JSStringRef targetNameRef = JSStringCreateWithUTF8CString(targetName);
		JSValueRef subObject = JSObjectGetProperty(context, base, targetNameRef, exception);
		JSStringRelease(targetNameRef);
		if(!JSValueIsUndefined(context, subObject))
		{
			return subObject;
		}
		return NULL;
	}

}

JSCustomEventDispatcher::JSCustomEventDispatcher() : JSEventDispatcher("SystemDispatcher") {
	isDispatching = false;
}

JSCustomEventDispatcher::~JSCustomEventDispatcher(){
}

bool JSCustomEventDispatcher::AddEventListener(JSEventListener *js_listener, std::string event_type)
{
	if(isDispatching)
	{
		//If this function is called during a callback, the action must be queued up to prevent the event_listener_map from getting changed
		add_list.push_back(ListenerEventPair(js_listener, event_type));
		return true;
	}
	else
	{
		return JSEventDispatcher::AddEventListener(js_listener, event_type);
	}
}

void JSCustomEventDispatcher::RemoveEventListener(JSEventListener *js_listener, std::string event_type)
{
	if(isDispatching)
	{
		//If this function is called during a callback, the action must be queued up to prevent the event_listener_map from getting changed
		remove_list.push_back(ListenerEventPair(js_listener, event_type));
	}
	else
	{
		JSEventDispatcher::RemoveEventListener(js_listener, event_type);
	}
}

void JSCustomEventDispatcher::RemoveAllListeners(JSEventListener *js_listener)
{
	if(isDispatching)
	{
		//If this function is called during a callback, the action must be queued up to prevent the event_listener_map from getting changed
		clear_list.push_back(js_listener);
	}
	else
	{
		JSEventDispatcher::RemoveAllListeners(js_listener);
	}
}

void JSCustomEventDispatcher::DispatchEvent(JSEvent js_event)
{
	//This flag tells the other functions to wait until we have finished processing all the listener callbacks
	isDispatching = true;
	JSEventDispatcher::DispatchEvent(js_event);
	isDispatching = false;
	processQueuedRequests();
}

void JSCustomEventDispatcher::DispatchEvent(JSEvent js_event, std::string event_string)
{
	//This flag tells the other functions to wait until we have finished processing all the listener callbacks
	isDispatching = true;
	JSEventDispatcher::DispatchEvent(js_event, event_string);
	isDispatching = false;
	processQueuedRequests();
}

void JSCustomEventDispatcher::processQueuedRequests()
{
	//Now we perform all the add/remove actions that were queued up during event callbacks
	typedef std::vector<JSEventListener *>::iterator L;
	for(L l = this->clear_list.begin(); l != this->clear_list.end(); ++l)
	{
		JSEventDispatcher::RemoveAllListeners(*l);
	}
	this->clear_list.clear();

	typedef std::vector<ListenerEventPair>::iterator K;
	for(K k = this->remove_list.begin(); k != this->remove_list.end(); ++k)
	{
		JSEventDispatcher::RemoveEventListener(k->js_listener, k->event_type);
	}
	this->remove_list.clear();

	for(K k = this->add_list.begin(); k != this->add_list.end(); ++k) {
		JSEventDispatcher::AddEventListener(k->js_listener, k->event_type);
	}
	this->add_list.clear();
}